# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from typing import List, Tuple, Any, Dict

import tensorflow as tf
from tensorflow import Tensor
from tensorflow_model_optimization.python.core.quantization.keras.quantizers import Quantizer
from packaging import version

# As from Tensorflow 2.6, keras is a separate package and some classes should be imported differently.
if version.parse(tf.__version__) < version.parse("2.6"):
    from tensorflow.python.keras.layers import Layer
else:
    from keras.engine.base_layer import Layer
from tensorflow.python.training.tracking.data_structures import ListWrapper
from tensorflow_model_optimization.python.core.quantization.keras.quantize_config import QuantizeConfig


class WeightsQuantizeConfig(QuantizeConfig):
    """
    QuantizeConfig to quantize a layer's weights.
    """

    def __init__(self,
                 w_quantizer: Quantizer,
                 weight_attrs: List[str] = None):
        """

        Args:
            w_quantizer: Quantizer for weights.
            weight_attrs: Weights attributes to quantize.
        """

        self.weight_attrs = weight_attrs
        self.w_quantizer = w_quantizer

    def get_config(self) -> Dict[str, Any]:
        """

        Returns: Configuration of WeightsQuantizeConfig

        """
        return {'w_quantizer': self.w_quantizer,
                'weight_attrs': self.weight_attrs}

    def get_weights_and_quantizers(self, layer: Layer) -> List[Tuple[Tensor, Any]]:
        """
        Get the layer's weights to quantize and quantizers.

        Args:
            layer: Layer wrapped with this WeightsQuantizeConfig

        Returns:
            List of weights and quantizers to quantize these weights.
        """
        return [(getattr(layer, self.weight_attrs[i]),
                 self.w_quantizer) for i in range(len(self.weight_attrs))]

    def get_activations_and_quantizers(self, layer: Layer) -> list:
        # For configurable activations we use get_output_quantizers,
        # Therefore, we do not need to implement this method.
        return []

    def set_quantize_weights(self, layer: Layer, quantize_weights: List[Tensor]):
        """
        Set layer's weights with quantized weights.

        Args:
            layer: Layer wrapped with this WeightsQuantizeConfig
            quantize_weights: Quantized weights to set to the layer

        Returns:
            None
        """
        if len(self.weight_attrs) != len(quantize_weights):
            raise ValueError(
                '`set_quantize_weights` called on layer {} with {} '
                'weight parameters, but layer expects {} values.'.format(
                    layer.name, len(quantize_weights), len(self.weight_attrs)))

        for weight_attr, weight in zip(self.weight_attrs, quantize_weights):
            current_weight = getattr(layer, weight_attr)
            if current_weight.shape != weight.shape:
                raise ValueError('Existing layer weight shape {} is incompatible with'
                                 'provided weight shape {}'.format(
                    current_weight.shape, weight.shape))

            setattr(layer, weight_attr, weight)

    def set_quantize_activations(self, layer, quantize_activations: ListWrapper):
        pass

    def get_output_quantizers(self, layer: Layer) -> List[Quantizer]:
        return []



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import copy

from collections.abc import Callable
from typing import Any

from model_compression_toolkit import QuantizationConfig
from model_compression_toolkit.core.common.quantization.node_quantization_config import NodeWeightsQuantizationConfig, \
    BaseNodeQuantizationConfig
from model_compression_toolkit.core.common.target_platform import QuantizationMethod, OpQuantizationConfig
from enum import Enum

IS_WEIGHTS = "is_weights"
WEIGHTS_QUANTIZATION_METHOD = "weights_quantization_method"


def transform_enum(v: Any):
    """
    If an enum is received it value is return otherwise the input is returned.
    Args:
        v: Any type

    Returns: Any

    """
    if isinstance(v, Enum):
        return v.value
    return v


def config_serialization(quantization_config: BaseNodeQuantizationConfig):
    """
    This function change BaseNodeQuantizationConfig to a dictionary
    Args:
        quantization_config: A BaseNodeQuantizationConfig for serialization

    Returns: A config dictionary of BaseNodeQuantizationConfig

    """
    config_data = {k: transform_enum(v) for k, v in quantization_config.__dict__.items() if
                   v is not isinstance(v, Callable)}
    config_data[IS_WEIGHTS] = isinstance(quantization_config, NodeWeightsQuantizationConfig)
    return config_data


def config_deserialization(in_config: dict) -> BaseNodeQuantizationConfig:
    """
    This function change config dictionary to it BaseNodeQuantizationConfig.
    Args:
        in_config:  A config dictionary of BaseNodeQuantizationConfig

    Returns: A BaseNodeQuantizationConfig

    """
    in_config = copy.deepcopy(in_config)
    qc = QuantizationConfig()
    op_cfg = OpQuantizationConfig(QuantizationMethod.POWER_OF_TWO, QuantizationMethod.POWER_OF_TWO,
                                  8, 8, True, True, True, True, 0, 0, 8)
    if in_config[IS_WEIGHTS]:
        nwqc = NodeWeightsQuantizationConfig(qc=qc,
                                             op_cfg=op_cfg,
                                             weights_quantization_fn=None,
                                             weights_quantization_params_fn=None,
                                             weights_channels_axis=0)
        in_config[WEIGHTS_QUANTIZATION_METHOD] = QuantizationMethod(in_config[WEIGHTS_QUANTIZATION_METHOD])

        nwqc.__dict__.update(in_config)
        return nwqc
    else:
        raise NotImplemented



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import tensorflow as tf
import numpy as np

from model_compression_toolkit import GumbelConfig
from model_compression_toolkit.gptq.keras.quantizer import quant_utils as qutils
from model_compression_toolkit.gptq.keras.quantizer.gumbel_rounding.base_gumbel_rounding import GumbelRoundingBase, \
    init_aux_var
from tensorflow_model_optimization.python.core.quantization.keras.quantize_wrapper import QuantizeWrapper
from tensorflow.python.framework.tensor_shape import TensorShape
from model_compression_toolkit.core.common.defaultdict import DefaultDict
from typing import Dict, Any, List
from model_compression_toolkit.gptq.keras.quantizer.gumbel_rounding.gumbel_softmax import gumbel_softmax, ste_gumbel
from model_compression_toolkit.core.common.constants import RANGE_MIN, RANGE_MAX
from model_compression_toolkit.gptq.common import gptq_constants
from model_compression_toolkit.gptq.keras.quantizer.ste_rounding.uniform_ste import rounding_uniform_quantizer


def gumbel_rounding_uniform_quantizer(tensor_data: tf.Tensor,
                                      auxvar_tensor: tf.Variable,
                                      range_min: tf.Tensor,
                                      range_max: tf.Tensor,
                                      n_bits: int) -> tf.Tensor:
    """
    Quantize a tensor according to given range (min, max) and number of bits.

    Args:
        tensor_data: Tensor values to quantize.
        auxvar_tensor: Tensor that manifests the bit shift the weight due to gptq.
        range_min: minimum bound of the range for quantization (or array of min values per channel).
        range_max: maximum bound of the range for quantization (or array of max values per channel).
        n_bits: Number of bits to quantize the tensor.

    Returns:
        Quantized data.
    """

    # adjusts the quantization rage so the quantization grid include zero.
    a, b = qutils.fix_range_to_include_zero(range_min, range_max, n_bits)

    # Compute the step size of quantized values.
    delta = (b - a) / (2 ** n_bits - 1)

    input_tensor_int = tf.stop_gradient(tf.floor((tensor_data - a) / delta))  # Apply rounding
    tensor_q = input_tensor_int + auxvar_tensor

    # Clip data in range
    clipped_tensor = qutils.ste_clip(tensor_q, min_val=0, max_val=2 ** n_bits - 1)

    # Quantize the data between min/max of quantization range.
    q = delta * clipped_tensor + a
    return q


class UniformGumbelRounding(GumbelRoundingBase):
    """
    Trainable constrained quantizer to quantize a layer inputs.
    """
    PTQ_MIN_RANGE = "_min_range"
    PTQ_MAX_RANGE = "_max_range"

    def __init__(self, num_bits: int, per_axis: bool, signed: bool, quantization_parameter_learning: bool,
                 min_range: np.ndarray, max_range: np.ndarray, gumbel_config: GumbelConfig,
                 quantization_axis: int = -1, max_lsbs_change_map: dict = DefaultDict({}, lambda: 1),
                 max_iteration: int = 10000):
        """
        Initialize a TrainableWeightQuantizer object with parameters to use
        for the quantization.

        Args:
            num_bits: Number of bits to use for the quantization.
            per_axis: Whether to quantize per-channel or per-tensor.
            signed: Signedness to use for the quantization range.
            quantization_parameter_learning: Threshold to use for the quantization.
            min_range: a numpy array of the min range.
            max_range: a numpy array of the max range.
            gumbel_config: A class with the gumbel rounding configurations.
            quantization_axis: Axis of tensor to use for the quantization.
            max_lsbs_change_map: a mapping between number of bits to max lsb change.
            max_iteration: The number of iteration of gptq.
        """
        super().__init__(num_bits, per_axis, signed, False, False, quantization_parameter_learning,
                         quantization_axis, gumbel_config,
                         max_lsbs_change_map,
                         max_iteration)
        self.threshold_shape = np.asarray(min_range).shape
        self.min_range = np.reshape(np.asarray(min_range), [-1]) if self.per_axis else float(
            min_range)
        self.max_range = np.reshape(np.asarray(max_range), [-1]) if self.per_axis else float(
            max_range)
        self.k_threshold = len(self.max_range) if self.per_axis else 1

    def build(self,
              tensor_shape: TensorShape,
              name: str,
              layer: QuantizeWrapper) -> Dict[str, tf.Variable]:
        """
        Add min and max variables to layer.
        Args:
            tensor_shape: Tensor shape the quantizer quantize.
            name: Prefix of variables names.
            layer: Layer to add the variables to. The variables are saved
            in the layer's scope.

        Returns:
            Dictionary of new variables.
        """
        super().build(tensor_shape, name, layer)

        if self.per_axis:
            input_shape = tensor_shape
            n_axis = len(input_shape)
            quantization_axis = n_axis + self.quantization_axis if self.quantization_axis < 0 else \
                self.quantization_axis
            reshape_shape = [self.k_threshold if i == quantization_axis else 1 for i in range(n_axis)]
        else:
            reshape_shape = [self.k_threshold]

        max_range = layer.add_weight(
            name + self.PTQ_MAX_RANGE,
            shape=reshape_shape,
            initializer=tf.keras.initializers.Constant(1.0),
            trainable=self.quantization_parameter_learning)
        max_range.assign(self.max_range.reshape(reshape_shape))

        min_range = layer.add_weight(
            name + self.PTQ_MIN_RANGE,
            shape=reshape_shape,
            initializer=tf.keras.initializers.Constant(1.0),
            trainable=self.quantization_parameter_learning)
        min_range.assign(self.min_range.reshape(reshape_shape))

        auxvar_tensor = layer.add_weight(
            name + gptq_constants.AUXVAR,
            shape=[self.m, *self.w_shape],
            initializer=tf.keras.initializers.Constant(0.0),
            trainable=True)
        w = getattr(layer.layer, name)

        q_error = w - rounding_uniform_quantizer(w, min_range, max_range,
                                                 n_bits=self.num_bits)
        ceil_indicator = (q_error < 0).numpy().astype("int")  # Negative error means the choose point is rounded to ceil.
        auxvar_tensor.assign(init_aux_var(ceil_indicator, self.w_shape, self.m))

        self.quantizer_parameters.update({gptq_constants.AUXVAR: auxvar_tensor,
                                          self.PTQ_MAX_RANGE: max_range,
                                          self.PTQ_MIN_RANGE: min_range})
        return self.quantizer_parameters

    def __call__(self, inputs: tf.Tensor,
                 training: bool,
                 weights: Dict[str, tf.Variable],
                 **kwargs: Dict[str, Any]):
        """
        Quantize a tensor.
        Args:
            inputs: Input tensor to quantize.
            training: Whether the graph is in training mode.
            weights: Dictionary of weights the quantizer can use to quantize the tensor.
            **kwargs: Additional variables the quantizer may receive.

        Returns:
            The quantized tensor.
        """

        auxvar = weights[gptq_constants.AUXVAR]
        ar_iter = weights[gptq_constants.GPTQ_ITER]
        ptq_min_range = weights[self.PTQ_MIN_RANGE]
        ptq_max_range = weights[self.PTQ_MAX_RANGE]
        aux_index_shift = weights[gptq_constants.AUXSHIFT]
        self.update_iteration(training, ar_iter)
        if self.per_axis:
            input_shape = inputs.shape
            n_axis = len(input_shape)
            quantization_axis = n_axis + self.quantization_axis if self.quantization_axis < 0 else \
                self.quantization_axis
            reshape_shape = [-1 if i == quantization_axis else 1 for i in range(n_axis)]

            reshape_shape_aux_ind = [-1, *[1 for _ in range(n_axis)]]
            #####################################################
            # Gumbel Softmax
            #####################################################
            if training:
                p_t = gumbel_softmax(auxvar, self.tau, self.g_t)
            else:
                p_t = gumbel_softmax(auxvar, self.minimal_temp, 0)
                p_t = ste_gumbel(p_t)
            self.p_t = p_t
            #####################################################
            # Calculate v hat and threshold hat
            #####################################################
            ptq_min_range = tf.reshape(ptq_min_range, reshape_shape)
            ptq_max_range = tf.reshape(ptq_max_range, reshape_shape)

            auxvar_hat = tf.reduce_sum(p_t * tf.reshape(aux_index_shift, reshape_shape_aux_ind), axis=0)
            #####################################################
            # Quantized Input
            #####################################################
            q_tensor = gumbel_rounding_uniform_quantizer(inputs, auxvar_hat,
                                                         ptq_min_range,
                                                         ptq_max_range,
                                                         self.num_bits)
            return q_tensor
        else:
            raise NotImplemented
            return gumbel_rounding_uniform_quantizer(inputs, auxvar_hat,
                                                     ptq_max_range,
                                                     ptq_min_range,
                                                     self.num_bits)

    def get_quant_config(self, layer) -> Dict[str, np.ndarray]:
        """
        Returns the config used to edit NodeQuantizationConfig after GPTQ retraining

        Args:
            layer: quantized layer

        Returns:
            A dictionary of attributes the quantize_config retraining has changed during GPTQ retraining.
            Keys must match NodeQuantizationConfig attributes

        """
        min_range = self.quantizer_parameters[self.PTQ_MIN_RANGE]
        max_range = self.quantizer_parameters[self.PTQ_MAX_RANGE]
        return {RANGE_MIN: min_range.numpy().reshape(self.threshold_shape),
                RANGE_MAX: max_range.numpy().reshape(self.threshold_shape)}

    def get_quantization_variable(self) -> List[tf.Tensor]:
        """
         This function return a list of quantizer parameters.
         Returns: A list of the quantizer parameters

         """
        return [self.quantizer_parameters[self.PTQ_MIN_RANGE], self.quantizer_parameters[self.PTQ_MAX_RANGE]]



from abc import abstractmethod
from typing import Any

from model_compression_toolkit import FrameworkInfo


class ModelValidation:
    """
    Class to define validation methods in order to validate the received model to quantize.
    """

    def __init__(self,
                 model: Any,
                 fw_info:FrameworkInfo):
        """
        Initialize a ModelValidation object.

        Args:
            model: Model to check its validity.
            fw_info: Information about the specific framework of the model.
        """
        self.model = model
        self.fw_info = fw_info

    @abstractmethod
    def validate_output_channel_consistency(self):
        """

        Validate that output channels index in all layers of the model are the same.
        If the model has layers with different output channels index, it should throw an exception.

        """
        raise NotImplemented(f'Framework validation class did not implement validate_output_channel_consistency')

    def validate(self):
        """

        Run all validation methods before the quantization process starts.

        """
        self.validate_output_channel_consistency()





# Copyright 2021 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import numpy as np


class BaseCollector(object):
    """
    Base class for statistics collection object.
    """

    def __init__(self):
        # When manipulation statistics in a granularity they were not collected by, the data is invalid.
        self.is_legal = True

    def scale(self, scale_factor: np.ndarray):
        """
        Scale all statistics in collector by some factor.
        Args:
            scale_factor: Factor to scale all collector's statistics by.

        """

        raise Exception(f'{self.__class__.__name__} needs to implement scale operation for its state.')

    def shift(self, shift_value: np.ndarray):
        """
        Shift all statistics in collector by some value.
        Args:
            shift_value: Value to shift all collector's statistics by.

        """

        raise Exception(f'{self.__class__.__name__} needs to implement shift operation for its state.')

    def update_legal_status(self, is_illegal: bool):
        """
        If statistics were manipulated in a granularity they were not collected by, the data is invalid,
        and its legal status should be tracked after each manipulation.
        Args:
            is_illegal: Whether current info is invalid or not.

        """

        self.is_legal = self.is_legal and not is_illegal

    def validate_data_correctness(self):
        """
        Verify the collector's statistics were manipulated in a granularity they were collected by.
        If the statistics are invalid, an exception is raised.
        """

        if not self.is_legal:
            raise Exception(f'{self.__class__.__name__} was manipulated per-channel,'
                            'but collected per-tensor. Data is invalid.')



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from typing import List
from enum import Enum

from model_compression_toolkit.core import common
from model_compression_toolkit.core.common.quantization.node_quantization_config import NodeWeightsQuantizationConfig, \
    NodeActivationQuantizationConfig, BaseNodeQuantizationConfig
from model_compression_toolkit.core.common.target_platform import QuantizationMethod


class QuantizationTarget(Enum):
    Activation = 0
    Weights = 1


class BaseQuantizer:
    def __init__(self,
                 quantization_config: BaseNodeQuantizationConfig,
                 quantization_target: QuantizationTarget,
                 quantization_method: List[QuantizationMethod]):
        """
        This class is a base quantizer which validate the the provide quantization config and define abstract function which any quantizer need to implment.

        Args:
            quantization_config: node quantization config class contins all the information above a quantizer.
            quantization_target: A enum which decided the qunaizer tensor type activation or weights.
            quantization_method: A list of "QuantizationMethod" enums which represent the quantizer supported methods.
        """
        self.quantization_config = quantization_config
        self.quantization_target = quantization_target
        self.quantization_method = quantization_method
        if self.quantization_target == QuantizationTarget.Weights:
            self.validate_weights()
            if self.quantization_config.weights_quantization_method not in quantization_method:
                common.Logger.error(
                    f'Quantization method mismatch expected:{quantization_method} and got  {self.quantization_config.weights_quantization_method}')
        elif self.quantization_target == QuantizationTarget.Activation:
            self.validate_activation()
            if self.quantization_config.activation_quantization_method not in quantization_method:
                common.Logger.error(
                    f'Quantization method mismatch expected:{quantization_method} and got  {self.quantization_config.activation_quantization_method}')
        else:
            common.Logger.error(
                f'Unknown Quantization Part:{quantization_target}')

    def initialize_quantization(self,
                                tensor_shape,
                                name: str,
                                layer):
        """
        This initializes the quantizer parameters given the parameter name and shape.

        Args:
            tensor_shape:  tensor shape
            name: tensor name
            layer: layer to quantized

        Returns: None

        """
        raise NotImplemented

    def __call__(self,
                 input2quantize,
                 training: bool):
        """
        Quantize a tensor.

        Args:
            input2quantize: Input tensor to quantize.
            training: Whether the graph is in training mode.

        Returns:
            The quantized tensor.
        """
        raise NotImplemented

    def activation_quantization(self) -> bool:
        """

        Returns: A boolean stating is this activation quantizer

        """
        return isinstance(self.quantization_config, NodeActivationQuantizationConfig)

    def weights_quantization(self) -> bool:
        """

        Returns: A boolean stating is this weights quantizer

        """
        return isinstance(self.quantization_config, NodeWeightsQuantizationConfig)

    def validate_weights(self) -> None:
        """
        This function valid the quantize config compare with it parameters.


        """
        if self.activation_quantization() or not self.weights_quantization():
            common.Logger.error(f'Expect weight quantization got activation')

    def validate_activation(self) -> None:
        """
        This function valid the quantize config compare with it parameters.

        """
        if not self.activation_quantization() or self.weights_quantization():
            common.Logger.error(f'Expect activation quantization got weight')



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================\
import numpy as np
from typing import Dict, Any

from model_compression_toolkit.core.common.constants import RANGE_MIN, RANGE_MAX
from model_compression_toolkit.core.common.target_platform import QuantizationMethod
from model_compression_toolkit.core.pytorch.quantizer.fake_quant_builder import uniform_quantization


class FakeQuantQuantizer:
    """
    Quantizer using TensorFlow fake quant layer to quantize activations.
    """

    def __init__(self,
                 nbits: int,
                 min_range: np.ndarray,
                 max_range: np.ndarray,
                 quantization_method: QuantizationMethod):
        """

        Args:
            nbits: Number of bits to quantize.
            min_range: Min quantization range.
            max_range: Max quantization range.
            quantization_method: Quantization method that is used (POT, Uniform, etc.)

        """
        self.nbits = nbits
        self.min_range = min_range
        self.max_range = max_range
        self.quantization_method = quantization_method

    def __call__(self, inputs):
        """
        Apply quantization to the input tensor.

        Args:
            inputs: Input tensor to be quantized.

        Returns:
            Quantized tensor.
        """
        return uniform_quantization(self.nbits,
                                    {RANGE_MIN: self.min_range,
                                     RANGE_MAX: self.max_range})(inputs)



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


from typing import Dict
from model_compression_toolkit.core import common
from model_compression_toolkit.core.common.framework_info import FrameworkInfo
from model_compression_toolkit import qunatizers_infrastructure as qi
from model_compression_toolkit.qat.keras.quantizer.ste_rounding.symmetirc_ste import STEWeightQuantizer
from model_compression_toolkit.qat.keras.quantizer.ste_rounding.uniform_ste import STEUniformWeightQuantizer

METHOD2QUANTIZER = {qi.QuantizationMethod.SYMMETRIC: STEWeightQuantizer,
                    qi.QuantizationMethod.POWER_OF_TWO: STEWeightQuantizer,
                    qi.QuantizationMethod.UNIFORM: STEUniformWeightQuantizer}


def quantization_dispatcher_builder(n: common.BaseNode,
                                    fw_info: FrameworkInfo,
                                    method2quantizer: Dict[
                                        qi.QuantizationMethod, qi.BaseKerasQuantizer] = METHOD2QUANTIZER) -> qi.KerasNodeQuantizationDispatcher:
    """
    Build a NodeQuantizationDispatcher for a node according to its quantization configuration and
    a global NoOpQuantizeConfig object.

    Args:
        n: Node to build its QuantizeConfig.
        fw_info: Framework information (e.g., mapping from layers to their attributes to quantize).
        method2quantizer: A mapping between quantization method to quantizer.

    Returns:
        A QuantizeConfig object with the appropriate quantizers (according to the node's
        quantization configuration).
    """
    nqd = qi.KerasNodeQuantizationDispatcher()
    if n.is_weights_quantization_enabled():
        attributes = fw_info.get_kernel_op_attributes(n.type)
        for attr in attributes:
            qunatizer_class = method2quantizer.get(n.final_weights_quantization_cfg.weights_quantization_method)
            if qunatizer_class is None:
                common.Logger.error(
                    f'Unknown Quantiztion method: {n.final_weights_quantization_cfg.weights_quantization_method}')
            nqd.add_weight_quantizer(attr, qunatizer_class(n.final_weights_quantization_cfg))

    return nqd



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from typing import Dict, Tuple, Any

from keras.engine.base_layer import Layer
from keras.layers import TFOpLambda
from tensorflow import TensorShape
from tensorflow_model_optimization.python.core.quantization.keras.quantize_config import QuantizeConfig
from tensorflow_model_optimization.python.core.quantization.keras.quantize_wrapper import QuantizeWrapper


class ExtendedQuantizeWrapper(QuantizeWrapper):

    """Quantizes the weights and activations of the keras layer it wraps, according
    to the quantization config that is passed. This class was created to deal with TFOpLambda that can
    not use TF QuantizeWrapper since it does not implement compute_output_shape.
    Notice that reused layers do not have a compute_output_shape method, thus the added method
    is irrelevant for wrapping them."""

    def __init__(self,
                 layer:Layer,
                 quantize_config:QuantizeConfig,
                 output_shape:Tuple[Any]=None,
                 **kwargs:Dict[str, Any]):
        """
        Create a wrapper for a keras layer.

        Args:
            layer: The keras layer to be quantized.
            quantize_config: `QuantizeConfig` to quantize the layer.
            output_shape: The output shape of the layer.
            **kwargs: Keyword arguments to build the base class QuantizeWrapper.
        """

        # TFOpLambda does not implement the method compute_output_shape which is mandatory for cloning the model
        # and use TF transformations. For this reason, we add the output_shape in the layer configuration and
        # add an implementation for compute_output_shape.
        self._output_shape = output_shape
        if isinstance(layer, TFOpLambda):
            layer.compute_output_shape = self._compute_output_shape

        super(ExtendedQuantizeWrapper, self).__init__(layer=layer,
                                                      quantize_config=quantize_config,
                                                      **kwargs)

    def _compute_output_shape(self, input_shape: TensorShape) -> Tuple[Any]:
        """
        Internal method that returns the output shape of the layer.

        Args:
            input_shape: Input shape the layer is expecting to have.

        Returns:
            The layer's output shape.
        """

        return self._output_shape

    def get_config(self) -> Dict[str, Any]:
        """

        Returns: The layer configuration with the output shape, so it can be deserialized.

        """
        cfg = super(ExtendedQuantizeWrapper, self).get_config()
        cfg.update({'output_shape': self._output_shape})
        return cfg




# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


import itertools
import pprint
from typing import List, Any, Dict, Tuple

from model_compression_toolkit.core.common.logger import Logger
from model_compression_toolkit.core.common.target_platform.targetplatform2framework.operations_to_layers import \
    OperationsToLayers, OperationsSetToLayers
from model_compression_toolkit.core.common.target_platform.targetplatform2framework.target_platform_capabilities_component import TargetPlatformCapabilitiesComponent
from model_compression_toolkit.core.common.target_platform.targetplatform2framework.layer_filter_params import LayerFilterParams
from model_compression_toolkit.core.common.immutable import ImmutableClass
from model_compression_toolkit.core.common.graph.base_node import BaseNode
from model_compression_toolkit.core.common.target_platform.op_quantization_config import QuantizationConfigOptions, \
    OpQuantizationConfig
from model_compression_toolkit.core.common.target_platform.operators import OperatorsSet, OperatorsSetBase
from model_compression_toolkit.core.common.target_platform.target_platform_model import TargetPlatformModel
from model_compression_toolkit.core.common.target_platform.targetplatform2framework.current_tpc import _current_tpc


class TargetPlatformCapabilities(ImmutableClass):
    """
    Attach framework information to a modeled hardware.
    """
    def __init__(self,
                 tp_model: TargetPlatformModel,
                 name: str = "base",
                 version: str = None):
        """

        Args:
            tp_model (TargetPlatformModel): Modeled hardware to attach framework information to.
            name (str): Name of the TargetPlatformCapabilities.
            version (str): TPC version.
        """

        super().__init__()
        self.name = name
        assert isinstance(tp_model, TargetPlatformModel), f'Target platform model that was passed to TargetPlatformCapabilities must be of type TargetPlatformModel, but has type of {type(tp_model)}'
        self.tp_model = tp_model
        self.op_sets_to_layers = OperationsToLayers() # Init an empty OperationsToLayers
        self.layer2qco, self.filterlayer2qco = {}, {} # Init empty mappings from layers/LayerFilterParams to QC options
        # Track the unused opsets for warning purposes.
        self.__tp_model_opsets_not_used = [s.name for s in tp_model.operator_set]
        self.remove_fusing_names_from_not_used_list()
        self.version = version

    def get_layers_by_opset_name(self, opset_name: str) -> List[Any]:
        """
        Get a list of layers that are attached to an OperatorsSet by the OperatorsSet name.

        Args:
            opset_name: OperatorsSet name to get its layers.


        Returns:
            List of layers/LayerFilterParams that are attached to the opset name.
        """
        opset = self.tp_model.get_opset_by_name(opset_name)
        if opset is None:
            Logger.warning(f'{opset_name} was not found in TargetPlatformCapabilities.')
            return None
        return self.get_layers_by_opset(opset)

    def get_layers_by_opset(self, op: OperatorsSetBase) -> List[Any]:
        """
        Get a list of layers that are attached to an OperatorsSet by the OperatorsSet object.

        Args:
            op: OperatorsSet object to get its layers.

        Returns:
            List of layers/LayerFilterParams that are attached to the OperatorSet object.
        """
        return self.op_sets_to_layers.get_layers_by_op(op)

    def get_fusing_patterns(self) -> List[List[Any]]:
        """

        Returns: List of patterns of layers/LayerFilterParams to fuse.

        """
        res = []
        for p in self.tp_model.fusing_patterns:
            ops = [self.get_layers_by_opset(x) for x in p.operator_groups_list]
            res.extend(itertools.product(*ops))
        return [list(x) for x in res]


    def get_info(self) -> Dict[str, Any]:
        """

        Returns: Summarization of information in the TargetPlatformCapabilities.

        """
        return {"Target Platform Capabilities": self.name,
                "Version": self.version,
                "Target Platform Model": self.tp_model.get_info(),
                "Operations to layers": {op2layer.name:[l.__name__ for l in op2layer.layers] for op2layer in self.op_sets_to_layers.op_sets_to_layers}}

    def show(self):
        """

        Display the TargetPlatformCapabilities.

        """
        pprint.pprint(self.get_info(), sort_dicts=False, width=110)

    def append_component(self, tpc_component: TargetPlatformCapabilitiesComponent):
        """
        Append a Component (like OperationsSetToLayers) to the TargetPlatformCapabilities.

        Args:
            tpc_component: Component to append to TargetPlatformCapabilities.

        """
        if isinstance(tpc_component, OperationsSetToLayers):
            self.op_sets_to_layers += tpc_component
        else:
            raise Exception(f'Trying to append an unfamiliar TargetPlatformCapabilitiesComponent of type: {type(tpc_component)}')

    def __enter__(self):
        """
        Init a TargetPlatformCapabilities object.
        """
        _current_tpc.set(self)
        return self

    def __exit__(self, exc_type, exc_value, tb):
        """
        Finalize a TargetPlatformCapabilities object.
        """
        if exc_value is not None:
            print(exc_value, exc_value.args)
            raise exc_value
        self.raise_warnings()
        self.layer2qco, self.filterlayer2qco = self._get_config_options_mapping()
        _current_tpc.reset()
        self.initialized_done()
        return self

    def get_default_op_qc(self) -> OpQuantizationConfig:
        """

        Returns: The default OpQuantizationConfig of the TargetPlatformModel that is attached
        to the TargetPlatformCapabilities.

        """
        return self.tp_model.get_default_op_quantization_config()

    def get_qco_by_node(self,
                        node: BaseNode) -> QuantizationConfigOptions:
        """
        Get the QuantizationConfigOptions of a node in a graph according
        to the mappings from layers/LayerFilterParams to the OperatorsSet in the TargetPlatformModel.

        Args:
            node: Node from graph to get its QuantizationConfigOptions.

        Returns:
            QuantizationConfigOptions of the node.
        """
        if node is None:
            raise Exception(f'Can not retrieve QC options for None node')
        for fl, qco in self.filterlayer2qco.items():
            if fl.match(node):
                return qco
        if node.type in self.layer2qco:
            return self.layer2qco.get(node.type)
        return self.tp_model.default_qco

    def _get_config_options_mapping(self) -> Tuple[Dict[Any, QuantizationConfigOptions],
                                                   Dict[LayerFilterParams, QuantizationConfigOptions]]:
        """
        Build mapping from layers to their QuantizationConfigOptions (and from LayerFilterParams
        to their QuantizationConfigOptions).

        Returns: Two mappings from layers/LayerFilterParams to their QuantizationConfigOptions.

        """
        layer2qco = {}
        filterlayer2qco = {}
        for op2layers in self.op_sets_to_layers.op_sets_to_layers:
            for l in op2layers.layers:
                qco = self.tp_model.get_config_options_by_operators_set(op2layers.name)
                if qco is None:
                    qco = self.tp_model.default_qco
                if isinstance(l, LayerFilterParams):
                    filterlayer2qco.update({l: qco})
                else:
                    layer2qco.update({l: qco})
        return layer2qco, filterlayer2qco


    def remove_fusing_names_from_not_used_list(self):
        """
        Remove OperatorSets names from the list of the unused sets (so a warning
        will not be displayed).
        """
        for f in self.tp_model.fusing_patterns:
            for s in f.operator_groups_list:
                self.remove_opset_from_not_used_list(s.name)

    def remove_opset_from_not_used_list(self,
                                        opset_to_remove: str):
        """
        Remove OperatorsSet name from the unused op list.

        Args:
            opset_to_remove: OperatorsSet name to remove.

        """
        if opset_to_remove in self.__tp_model_opsets_not_used:
            self.__tp_model_opsets_not_used.remove(opset_to_remove)

    def raise_warnings(self):
        """

        Log warnings regards unused opsets.

        """
        for op in self.__tp_model_opsets_not_used:
            Logger.warning(f'{op} is defined in TargetPlatformModel, but is not used in TargetPlatformCapabilities.')





# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Tuple
import numpy as np
import torch
from torch.nn import Conv2d
import torch.nn.functional as F
from model_compression_toolkit.core.pytorch.utils import to_torch_tensor, torch_tensor_to_numpy
from model_compression_toolkit.core.common.graph.graph_matchers import NodeOperationMatcher
from model_compression_toolkit.core.common import BaseNode
from model_compression_toolkit.core.common.substitutions.linear_collapsing import Conv2DCollapsing
from model_compression_toolkit.core.pytorch.constants import KERNEL, KERNEL_SIZE, STRIDES, DILATIONS, BIAS, USE_BIAS, FILTERS, PADDING, GROUPS
from model_compression_toolkit.core.common.logger import Logger


def linear_collapsing_node_matchers() -> Tuple[NodeOperationMatcher, NodeOperationMatcher]:
    """
    Function generates matchers for matching:
    (Conv2D, Conv2D)[activation=linear] -> Conv2D.
    Returns:
        Matcher for 2 consecutive linear convolution
    """
    first_node = NodeOperationMatcher(Conv2d)
    second_node = NodeOperationMatcher(Conv2d)
    return first_node, second_node


def conv2d_collapsing_fn(first_node: BaseNode,
                         second_node: BaseNode,
                         kernel_str: str,
                         bias_str: str) -> Tuple[np.ndarray, np.ndarray]:
    """
    Collapsing 2 convolutions to one convolution: Out = k2*(k1*x+b1)+b2 = k2*k1*x+k2*b1+b2 = k*x+b
    We calculate k=k2*k1 (collapsed kernel) by injecting identity tensor to the convolutions and extract the output
    We calculate b=k2*b1+b2 (collapsed bias) matrix multiplication
    Args:
        first_node: First layer node to collapse to second layer node
        second_node: Second layer node
        kernel_str: The framework specific attribute name of the convolution layer's weight/kernel.
        bias_str: The framework specific attribute name of the convolution layer's bias.
    Returns:
        The modified layer node's weights: kernel, bias
    """
    if first_node.type == Conv2d and second_node.type == Conv2d:
        # Get nodes attributes
        kernel1 = first_node.get_weights_by_keys(kernel_str)
        kernel2 = second_node.get_weights_by_keys(kernel_str)
        bias1 = first_node.get_weights_by_keys(bias_str)
        bias2 = second_node.get_weights_by_keys(bias_str)
        strides1 = first_node.framework_attr[STRIDES]
        strides2 = second_node.framework_attr[STRIDES]

        # --------------------------------------- #
        # Kernel collapsing: k=k2*k1
        # --------------------------------------- #
        # Inspired by https://arxiv.org/pdf/2103.09404.pdf - Algorithm1

        # Generate identity input with padding
        kx, ky = kernel1.shape[2] + kernel2.shape[2] - 1, kernel1.shape[3] + kernel2.shape[3] - 1
        x_pad, y_pad = 2 * kx - 1, 2 * ky - 1
        in_tensor = torch.eye(kernel1.shape[1])
        in_tensor = torch.unsqueeze(torch.unsqueeze(in_tensor, 2), 3)
        in_tensor = F.pad(in_tensor, (int(np.ceil((x_pad - 1) / 2)),
                                      int(np.floor((x_pad - 1) / 2)),
                                      int(np.ceil((y_pad - 1) / 2)),
                                      int(np.floor((y_pad - 1) / 2))))

        # Run first Conv2D
        conv1_out = F.conv2d(input=to_torch_tensor(in_tensor), weight=to_torch_tensor(kernel1), stride=strides1, padding=(0,0))

        # Run second Conv2D
        kernel2_torch = to_torch_tensor(kernel2)
        conv2_out = F.conv2d(input=conv1_out, weight=kernel2_torch, stride=strides2)

        # Extract collapsed kernel from output: the collapsed kernel is the output of the convolution after fixing the dimension
        kernel_collapsed = torch_tensor_to_numpy(torch.permute(torch.flip(conv2_out,[3,2]), dims=[1,0,2,3]))

        # --------------------------------------- #
        # Bias collapsing: b=k2*b1+b2
        # --------------------------------------- #
        bias_collapsed = None
        if bias1 is not None:
            bias1_torch = to_torch_tensor(bias1)
            bias_collapsed = torch_tensor_to_numpy(torch.matmul(torch.sum(kernel2_torch,dim=(2, 3)), bias1_torch))
            if bias2 is not None:
                bias_collapsed += bias2
        elif bias2 is not None:
            bias_collapsed = bias2

        return kernel_collapsed, bias_collapsed
    else:
        Logger.error("No supported layer collapsing of {} and {}".format(first_node.type, second_node.type))


def pytorch_linear_collapsing() -> Conv2DCollapsing:
    """
    Returns:
        A Conv2DCollapsing initialized for pytorch models.
    """
    first_node, second_node = linear_collapsing_node_matchers()
    return Conv2DCollapsing(first_node,
                            second_node,
                            conv2d_collapsing_fn,
                            KERNEL,
                            KERNEL_SIZE,
                            BIAS,
                            USE_BIAS,
                            STRIDES,
                            PADDING,
                            DILATIONS,
                            GROUPS,
                            FILTERS)



from model_compression_toolkit.qunatizers_infrastructure.common.base_quantizer import QuantizationTarget, \
    QuantizationMethod

from model_compression_toolkit.qunatizers_infrastructure.keras.quantize_wrapper import KerasQuantizationWrapper
from model_compression_toolkit.qunatizers_infrastructure.keras.base_keras_quantizer import BaseKerasQuantizer
from model_compression_toolkit.qunatizers_infrastructure.keras.load_model import keras_load_quantized_model
from model_compression_toolkit.qunatizers_infrastructure.keras.keras_node_quantization_dispatcher import \
    KerasNodeQuantizationDispatcher



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import torch
import torch.nn as nn
from typing import List, Union
from model_compression_toolkit.gptq.common.gptq_config import GradientPTQConfig
from model_compression_toolkit.gptq.pytorch.quantizer.gumbel_rounding.base_gumbel_weights_quantizer import BaseGumbelWeightQuantizer, init_aux_var
from model_compression_toolkit.core.pytorch.utils import to_torch_tensor, torch_tensor_to_numpy
from model_compression_toolkit.gptq.pytorch.quantizer.quant_utils import symmetric_quantizer
from model_compression_toolkit.gptq.pytorch.quantizer.quant_utils import ste_clip, ste_gumbel, gumbel_softmax, power_of_two_max
from model_compression_toolkit.gptq.common.gptq_constants import AUXVAR, THRESHOLD_TENSOR, TEMP, SCALE_TENSOR
from model_compression_toolkit.core.common.quantization.node_quantization_config import NodeWeightsQuantizationConfig
from model_compression_toolkit.core.common.constants import THRESHOLD


class SymmetricGumbelWeightQuantizer(BaseGumbelWeightQuantizer):
    """
    Class that implements a quantizer with trainable parameters to be used for GPTQ training.
    """

    def __init__(self,
                 weights_quantization_cfg: NodeWeightsQuantizationConfig,
                 gptq_config: GradientPTQConfig,
                 weight: torch.Tensor):
        """
        Construct a Pytorch model that utilize a fake weight quantizer of Symmetric Gumbel rounding
        Args:
            weights_quantization_cfg: Configuration of weight quantization
            gptq_config: GradientPTQConfig object with parameters about the tuning process.
            weight: weight for auxiliary tensor creation.
        """
        super().__init__(weights_quantization_cfg, gptq_config, weight.shape)
        self.signed = True
        self.min_int = -int(self.signed) * (2 ** (self.num_bits - int(self.signed)))
        self.max_int = (2 ** (self.num_bits - int(self.signed))) - 1
        self.threshold_tensor = to_torch_tensor(weights_quantization_cfg.weights_quantization_params.get(THRESHOLD))
        self.scale_tensor = torch.ones(self.weight_shape)

        # Set trainable tensors
        self.set_trainable_params(weight)


    def set_trainable_params(self, weight: torch.nn.Parameter):
        """
        A function to set a list of trainable parameters of the quantizer for GPTQ retraining
        Args:
            weight: weight for auxiliary tensor creation.
        """
        q_error = weight - symmetric_quantizer(weight,
                                               self.threshold_tensor,
                                               num_bits=self.num_bits,
                                               signed=True,
                                               power_of_two=self.power_of_two)
        ceil_indicator = (q_error < 0).int()  # Negative error means the choosen point is rounded to ceil.
        self.aux_tensor = nn.Parameter(to_torch_tensor(init_aux_var(ceil_indicator, self.weight_shape, self.m)), requires_grad=True)
        self.trainable_params.update({AUXVAR: self.aux_tensor})
        self.temp_tensor = nn.Parameter(to_torch_tensor(self.maximal_temp*torch.ones([1,*self.weight_shape])), requires_grad=True)
        self.trainable_params.update({TEMP: self.temp_tensor})
        if self.quantization_parameter_learning and not self.power_of_two:
            self.scale_tensor = nn.Parameter(to_torch_tensor(self.scale_tensor), requires_grad=True)
            self.trainable_params.update({SCALE_TENSOR: self.scale_tensor})
        elif self.quantization_parameter_learning:
            self.threshold_tensor = nn.Parameter(self.threshold_tensor, requires_grad=True)
            self.trainable_params.update({THRESHOLD_TENSOR: self.threshold_tensor})
        else:
            self.trainable_params.update({THRESHOLD_TENSOR: self.threshold_tensor})

    def get_aux_variable(self) -> torch.Tensor:
        """
        Returns auxiliary trainable variables
        """
        return self.trainable_params.get(AUXVAR)

    def get_quantization_variable(self) -> Union[torch.Tensor, List]:
        """
        Returns quantization trainable variables
        """
        if self.quantization_parameter_learning and not self.power_of_two:
            return [self.trainable_params.get(SCALE_TENSOR)]
        else:
            return [self.trainable_params.get(THRESHOLD_TENSOR)]

    def get_temperature_variable(self) -> Union[torch.Tensor, List]:
        """
        Returns temperature trainable variables
        """
        return self.trainable_params.get(TEMP)

    def get_weight_quant_params(self) -> dict:
        """
        Returns weight quantization dictionary params
        """
        threshold_tensor = self.threshold_tensor
        if self.power_of_two:
            threshold_tensor = power_of_two_max(threshold_tensor)
        elif self.quantization_parameter_learning:
            threshold_tensor = threshold_tensor*self.scale_tensor
        return {THRESHOLD: torch_tensor_to_numpy(threshold_tensor.detach())}

    def forward(self, w: nn.Parameter, training:bool = True) -> nn.Parameter:
        """
        Weight fake quantizer
        Args:
            w: weights to quantize.
            training: whether in training mode or not
        Returns:
            quantized weights
        """
        self.update_iteration(training)

        #####################################################
        # Gumbel Softmax
        #####################################################
        if training:
            gumbel_scale = self.gumbel_scale if self.gumbel_scale_per_bitwidth is None \
                else self.gumbel_scale_per_bitwidth.get(self.num_bits, self.gumbel_scale)
            self.p_t = gumbel_softmax(self.aux_tensor, self.tau, self.g_t, gumbel_scale=gumbel_scale)
        else:
            self.p_t = ste_gumbel(gumbel_softmax(self.aux_tensor, self.minimal_temp, 0))

        auxhat_tensor = torch.sum(self.p_t * self.shift_tensor.reshape(self.reshape_aux_shift), dim=0)

        #####################################################
        # Quantizer
        #####################################################
        threshold_tensor = self.threshold_tensor
        if self.power_of_two:
            threshold_tensor = power_of_two_max(threshold_tensor)
        delta_tensor = threshold_tensor / (2 ** (self.num_bits-int(self.signed)))
        w0 = torch.floor(w / delta_tensor).detach()
        w1 = w0 + auxhat_tensor
        w2 = ste_clip(w1, min_val=self.min_int, max_val=self.max_int)
        w_q = delta_tensor * w2
        # Scale
        if self.quantization_parameter_learning and not self.power_of_two:
            w_q *= self.scale_tensor
        return w_q




# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from model_compression_toolkit.core.common import Logger
from model_compression_toolkit.core.common.constants import FOUND_TF

if FOUND_TF:
    import tensorflow as tf
    from model_compression_toolkit import qunatizers_infrastructure as qi
    from model_compression_toolkit.qunatizers_infrastructure.keras.base_keras_quantizer import BaseKerasQuantizer

    keras = tf.keras


    def keras_load_quantized_model(filepath, custom_objects=None, compile=True, options=None):
        """
        This function wraps the keras load model and MCT quantization custom class to it.

        Args:
            filepath: the model file path.
            custom_objects: Additional custom objects
            compile: Boolean, whether to compile the model after loading.
            options: Optional `tf.saved_model.LoadOptions` object that specifies options for loading from SavedModel.

        Returns: A keras Model

        """
        qi_custom_objects = {subclass.__name__: subclass for subclass in BaseKerasQuantizer.__subclasses__()}
        qi_custom_objects.update({qi.KerasQuantizationWrapper.__name__: qi.KerasQuantizationWrapper,
                                  qi.KerasNodeQuantizationDispatcher.__name__: qi.KerasNodeQuantizationDispatcher})
        if custom_objects is not None:
            qi_custom_objects.update(custom_objects)
        return tf.keras.models.load_model(filepath,
                                          custom_objects=qi_custom_objects, compile=compile,
                                          options=options)
else:
    def keras_load_quantized_model(filepath, custom_objects=None, compile=True, options=None):
        """
        This function wraps the keras load model and MCT quantization custom class to it.

        Args:
            filepath: the model file path.
            custom_objects: Additional custom objects
            compile: Boolean, whether to compile the model after loading.
            options: Optional `tf.saved_model.LoadOptions` object that specifies options for loading from SavedModel.

        Returns: A keras Model

        """
        Logger.critical('Installing tensorflow and tensorflow_model_optimization is mandatory '
                        'when using keras_load_quantized_model. '
                        'Could not find Tensorflow package.')



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Any

from keras.engine.input_layer import InputLayer
from tensorflow_model_optimization.python.core.quantization.keras.quantize_wrapper import QuantizeWrapperV2

from model_compression_toolkit.core.common import Logger
from model_compression_toolkit.exporter.model_wrapper.keras.builder.quantize_config_to_node import \
    SUPPORTED_QUANTIZATION_CONFIG
from model_compression_toolkit.exporter.model_wrapper.keras.extended_quantize_wrapper import ExtendedQuantizeWrapper


def is_keras_layer_exportable(layer: Any) -> bool:
    """
    Check whether a Keras layer is a valid exportable layer or not.

    Args:
        layer: Keras layer to check if considered to be valid for exporting.

    Returns:
        Check whether a Keras layer is a valid exportable layer or not.
    """
    # Keras Input layers are not wrapped
    if isinstance(layer, InputLayer):
        return True

    valid_layer = isinstance(layer, ExtendedQuantizeWrapper)
    if not valid_layer:
        Logger.error(f'Exportable layer must be wrapped using ExtendedQuantizeWrapper, but layer {layer.name} is of type {type(layer)}')

    valid_quantize_config = type(layer.quantize_config) in SUPPORTED_QUANTIZATION_CONFIG
    if not valid_quantize_config:
        Logger.error(f'QuantizeConfig of layer is not supported. Type: {type(layer.quantize_config)}. Supported configs: {SUPPORTED_QUANTIZATION_CONFIG}.')

    return True



# Copyright 2021 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


import numpy as np

from model_compression_toolkit.core.common.collectors.base_collector import BaseCollector
from model_compression_toolkit.core.common.constants import LAST_AXIS


class MeanCollector(BaseCollector):
    """
        Class to collect observed per channel mean values of tensors that goes through it (passed to update).
        The mean is calculated using a exponential moving average with bias correction.
    """

    def __init__(self,
                 axis: int,
                 beta: float = 0.99):
        """
        Instantiate a per channel mean collector using a exponential moving average with bias correction.

        Args:
            axis: Compute the mean with regard to this axis.
            beta: Parameter for mean smoothing by EMA.
        """
        super().__init__()
        self.axis = axis
        self.__state_internal = np.array([0.0])  # mean per-channel
        self.__state_internal_correction = None
        self.beta = beta
        self.i = 0.0

    def scale(self, scale_factor: np.ndarray):
        """
        Scale all statistics in collector by some factor.
        Since mean was collected per-channel, it can be scaled either by a single factor or a factor
        per-channel.
        The scaling is done using the corrected mean.

        Args:
            scale_factor: Factor to scale all collector's statistics by.

        """

        self.__state_internal_correction *= scale_factor

    def shift(self, shift_value: np.ndarray):
        """
        Shift all statistics in collector by some value.
        Since mean was collected per-channel, it can be shifted either by a single value or a
        shifting value per-channel.
        The shifting is done using the corrected mean.

        Args:
            shift_value: Value to shift all collector's statistics by.

        """

        self.__state_internal_correction += shift_value

    @property
    def state(self):
        """
        The mean is kept internal and corrected when accessed from outside the collector.

        Returns: Mean of the collector after bias correction.
        """
        self.validate_data_correctness()
        return self.__state_internal_correction

    def update(self,
               x: np.ndarray):
        """
        Update the mean using a new tensor x to consider.

        Args:
            x: Tensor that goes through the mean collector and needs to be considered in the mean computation.
        """

        self.i += 1  # Update the iteration index
        axis = (len(x.shape) - 1) if self.axis == LAST_AXIS else self.axis
        n = x.shape[axis]
        transpose_index = [axis, *[i for i in range(len(x.shape)) if i != axis]]
        mu = np.mean(np.reshape(np.transpose(x, transpose_index), [n, -1]), axis=-1)  # compute mean per channel
        update_state = self.beta * self.__state_internal + (1 - self.beta) * mu
        self.__state_internal = update_state

        # Since we use a weighted mean, initial values can be distorted,
        # so use bias correction to compensate it.
        bias_correction = 1 - (self.beta ** self.i)
        self.__state_internal_correction = self.__state_internal / bias_correction



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from model_compression_toolkit.core.common import BaseNode, Graph, BaseSubstitution
from model_compression_toolkit.core.common.logger import Logger
from model_compression_toolkit.core.common.graph.virtual_activation_weights_node import VirtualActivationWeightsNode


class BaseVirtualActivationWeightsComposition(BaseSubstitution):
    def __init__(self, matcher_instance):
        super().__init__(matcher_instance=matcher_instance)

    def substitute(self,
                   graph: Graph,
                   weights_node: BaseNode) -> Graph:
        """
        Combines an activation --> weights edge's node into one virtual composed node that contains the activation
        operation and the linear operation (in that order).
        The node's quantization configuration candidates include the cartesian product of both node's candidates.
        Note that if the activation node has multiple outputs (beside its matched weights node) than the substitution
        would not apply.

        Args:
            graph: Graph we apply the substitution on.
            weights_node: A node with linear operation to be combined with its preceding activation.

        Returns:
            Graph after applying the substitution.
        """

        predecessors = graph.get_prev_nodes(weights_node)
        if len(predecessors) != 1:
            return graph

        act_node = predecessors[0]

        if len(graph.out_edges(act_node)) > 1:
            Logger.warning(f"Node {act_node.name} has multiple outgoing edges, which is not supported with "
                           f"mixed-precision bit-operations KPI, thus, edge {act_node.name} --> {weights_node.name} "
                           f"would not be counted in the bit-operations calculations.")
            return graph

        # Virtual composed activation-weights node
        # we pass a dummy initialization dict to initialize the super BaseNode class,
        # the actual arguments values are irrelevant because they are being overridden or not used
        v_node = VirtualActivationWeightsNode(act_node,
                                              weights_node,
                                              **weights_node.__dict__)

        # Update graph
        graph.add_node(v_node)
        graph.reconnect_in_edges(current_node=act_node, new_node=v_node)
        graph.reconnect_out_edges(current_node=weights_node, new_node=v_node)
        graph.replace_input_node(current_node=act_node, new_node=v_node)
        graph.replace_output_node(current_node=weights_node, new_node=v_node)
        graph.remove_edge(act_node, weights_node)
        graph.remove_node(weights_node)
        graph.remove_node(act_node)

        return graph



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from typing import List, Tuple, Any, Dict

import tensorflow as tf
from tensorflow import Tensor
from tensorflow_model_optimization.python.core.quantization.keras.quantizers import Quantizer
from packaging import version

# As from Tensorflow 2.6, keras is a separate package and some classes should be imported differently.
if version.parse(tf.__version__) < version.parse("2.6"):
    from tensorflow.python.keras.layers import Layer
else:
    from keras.engine.base_layer import Layer
from tensorflow.python.training.tracking.data_structures import ListWrapper
from tensorflow_model_optimization.python.core.quantization.keras.quantize_config import QuantizeConfig


class ActivationQuantizeConfig(QuantizeConfig):
    """
    QuantizeConfig to quantize a layer's activations.
    """

    def __init__(self,
                 activation_quantizer: Quantizer):
        """

        Args:
            activation_quantizer: Quantizer for quantization the layer's activations.
        """

        self.activation_quantizer = activation_quantizer


    def get_config(self) -> Dict[str, Any]:
        """

        Returns: Configuration of ActivationQuantizeConfig

        """
        return {
            'activation_quantizer': self.activation_quantizer}

    def get_weights_and_quantizers(self, layer: Layer) -> List[Tuple[Tensor, Any]]:
        return []

    def get_activations_and_quantizers(self, layer: Layer) -> list:
        # For configurable activations we use get_output_quantizers,
        # Therefore, we do not need to implement this method.
        return []

    def set_quantize_weights(self, layer: Layer, quantize_weights: List[Tensor]):
        pass

    def set_quantize_activations(self, layer, quantize_activations: ListWrapper):
        pass

    def get_output_quantizers(self, layer: Layer) -> List[Quantizer]:
        """
        Quantize layer's outputs.

        Args:
            layer: Layer to quantize its activations.

        Returns: List of activation quantizers.

        """
        return [self.activation_quantizer]



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from enum import Enum
from typing import Callable

import torch.nn

from model_compression_toolkit.core.common import Logger
from model_compression_toolkit.exporter.model_exporter.pytorch.fakely_quant_onnx_pytorch_exporter import \
    FakelyQuantONNXPyTorchExporter
from model_compression_toolkit.exporter.model_exporter.pytorch.fakely_quant_torchscript_pytorch_exporter import \
    FakelyQuantTorchScriptPyTorchExporter


class PyTorchExportMode(Enum):
    FAKELY_QUANT_TORCHSCRIPT = 0
    FAKELY_QUANT_ONNX = 1


def pytorch_export_model(model: torch.nn.Module,
                         is_layer_exportable_fn: Callable,
                         mode: PyTorchExportMode = PyTorchExportMode.FAKELY_QUANT_TORCHSCRIPT,
                         save_model_path: str = None,
                         repr_dataset: Callable = None) -> None:
    """
    Prepare and return fully quantized model for export. Save exported model to
    a path if passed.

    Args:
        model: Model to export.
        is_layer_exportable_fn: Callable to check whether a layer can be exported or not.
        mode: Mode to export the model according to.
        save_model_path: Path to save the model.
        repr_dataset: Representative dataset for tracing the pytorch model (mandatory for exporting it).

    """

    if mode == PyTorchExportMode.FAKELY_QUANT_TORCHSCRIPT:
        exporter = FakelyQuantTorchScriptPyTorchExporter(model,
                                                         is_layer_exportable_fn,
                                                         save_model_path,
                                                         repr_dataset)

    elif mode == PyTorchExportMode.FAKELY_QUANT_ONNX:
        exporter = FakelyQuantONNXPyTorchExporter(model,
                                                  is_layer_exportable_fn,
                                                  save_model_path,
                                                  repr_dataset)

    else:
        Logger.critical(
            f'Unsupported mode was used {mode.name} to export PyTorch model. Please see API for supported modes.')

    exporter.export()



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Dict, Any, List

from model_compression_toolkit.core.common import Logger
from model_compression_toolkit.core.common.constants import FOUND_TF
from model_compression_toolkit.core.common.quantization.node_quantization_config import BaseNodeQuantizationConfig
from model_compression_toolkit.core.common.target_platform import QuantizationMethod

from model_compression_toolkit.qunatizers_infrastructure.common.base_quantizer import BaseQuantizer, QuantizationTarget

if FOUND_TF:
    QUANTIZATION_CONFIG = 'qunatization_config'
    from model_compression_toolkit.qunatizers_infrastructure.keras.config_serialization import config_serialization, \
        config_deserialization


    class BaseKerasQuantizer(BaseQuantizer):
        def __init__(self,
                     quantization_config: BaseNodeQuantizationConfig,
                     quantization_target: QuantizationTarget,
                     quantization_method: List[QuantizationMethod]):
            """
            This class is a base quantizer which validate the provide quantization config and define abstract function which any quantizer need to implment.
            This class add to the base quantizer get_config and from_config function to enable keras load and save model.
            Args:
                quantization_config: node quantization config class contins all the information above a quantizer.
                quantization_target: A enum which decided the qunaizer tensor type activation or weights.
                quantization_method: A list of enums which represent the quantizer supported methods.
            """
            super().__init__(quantization_config, quantization_target, quantization_method)

        def get_config(self) -> Dict[str, Any]:
            """

            Returns: Configuration of BaseKerasQuantizer.

            """
            return {QUANTIZATION_CONFIG: config_serialization(self.quantization_config)}

        @classmethod
        def from_config(cls, config: dict):
            """

            Args:
                config(dict): dictonory  of  BaseKerasQuantizer Configuration

            Returns: A BaseKerasQuantizer

            """
            config = config.copy()
            quantization_config = config_deserialization(config[QUANTIZATION_CONFIG])
            # Note that a quantizer only receive quantization config and the rest of define hardcoded inside the speficie quantizer.
            return cls(quantization_config=quantization_config)

else:
    class BaseKerasQuantizer(BaseQuantizer):
        def __init__(self, quantization_config: BaseNodeQuantizationConfig, quantization_target: QuantizationTarget,
                     quantization_method: List[QuantizationMethod]):
            super().__init__(quantization_config, quantization_target, quantization_method)
            Logger.critical('Installing tensorflow and tensorflow_model_optimization is mandatory '
                            'when using BaseKerasQuantizer. '
                            'Could not find Tensorflow package.')
